/* ---------------------------------------------------------------------------
    Copyright (c) 2003-2011 Micrel, Inc.
    Copyright (c) 2015 Microchip Technology Inc.
   ---------------------------------------------------------------------------

    NdisOid.c - NDIS driver OID query functions.

    Author      Date        Description
    THa         01/26/11    Ported from NDIS 5.0 driver with code from
                            Microsoft sample driver.
   ---------------------------------------------------------------------------
*/


#include "target.h"
#include "NdisDevice.h"


/* -------------------------------------------------------------------------- */

#define OID_DEVICE_MODE  0xff0100cc

// List of supported OID for this driver

extern NDIS_OID SupportedOids[];

/* -------------------------------------------------------------------------- */

#ifdef DBG
VOID DbgPrintOidName (
    IN  PCHAR    Method,
    IN  NDIS_OID Oid )
{
    PCHAR oidName = NULL;
    BOOLEAN ignore = FALSE;

    switch (Oid)
    {

        #undef MAKECASE
        #define MAKECASE(oidx) case oidx: oidName = #oidx "\n"; break;
        #define IGNORECASE(oidx) case oidx: ignore = TRUE; break;

        /* Operational OIDs */
        MAKECASE(OID_GEN_SUPPORTED_LIST)
        MAKECASE(OID_GEN_HARDWARE_STATUS)
        MAKECASE(OID_GEN_MEDIA_SUPPORTED)
        MAKECASE(OID_GEN_MEDIA_IN_USE)
        MAKECASE(OID_GEN_MAXIMUM_LOOKAHEAD)
        MAKECASE(OID_GEN_MAXIMUM_FRAME_SIZE)
        MAKECASE(OID_GEN_LINK_SPEED)
        MAKECASE(OID_GEN_TRANSMIT_BUFFER_SPACE)
        MAKECASE(OID_GEN_RECEIVE_BUFFER_SPACE)
        MAKECASE(OID_GEN_TRANSMIT_BLOCK_SIZE)
        MAKECASE(OID_GEN_RECEIVE_BLOCK_SIZE)
        MAKECASE(OID_GEN_VENDOR_ID)
        MAKECASE(OID_GEN_VENDOR_DESCRIPTION)
        MAKECASE(OID_GEN_VENDOR_DRIVER_VERSION)
        MAKECASE(OID_GEN_CURRENT_PACKET_FILTER)
        MAKECASE(OID_GEN_CURRENT_LOOKAHEAD)
        MAKECASE(OID_GEN_DRIVER_VERSION)
        MAKECASE(OID_GEN_MAXIMUM_TOTAL_SIZE)
        MAKECASE(OID_GEN_PROTOCOL_OPTIONS)
        MAKECASE(OID_GEN_MAC_OPTIONS)
        MAKECASE(OID_GEN_MEDIA_CONNECT_STATUS)
        MAKECASE(OID_GEN_MAXIMUM_SEND_PACKETS)
        MAKECASE(OID_GEN_SUPPORTED_GUIDS)
        MAKECASE(OID_GEN_NETWORK_LAYER_ADDRESSES)
        MAKECASE(OID_GEN_TRANSPORT_HEADER_OFFSET)
        MAKECASE(OID_GEN_MEDIA_CAPABILITIES)
        MAKECASE(OID_GEN_PHYSICAL_MEDIUM)
        MAKECASE(OID_GEN_MACHINE_NAME)
        MAKECASE(OID_GEN_VLAN_ID)
        MAKECASE(OID_GEN_RNDIS_CONFIG_PARAMETER)

        /* Operational OIDs for NDIS 6.0 */
        MAKECASE(OID_GEN_MAX_LINK_SPEED)
        MAKECASE(OID_GEN_LINK_STATE)
        MAKECASE(OID_GEN_LINK_PARAMETERS)
        MAKECASE(OID_GEN_MINIPORT_RESTART_ATTRIBUTES)
        MAKECASE(OID_GEN_ENUMERATE_PORTS)
        MAKECASE(OID_GEN_PORT_STATE)
        MAKECASE(OID_GEN_PORT_AUTHENTICATION_PARAMETERS)
        MAKECASE(OID_GEN_INTERRUPT_MODERATION)
        MAKECASE(OID_GEN_PHYSICAL_MEDIUM_EX)

        /* Statistical OIDs */
        IGNORECASE(OID_GEN_XMIT_OK)
        IGNORECASE(OID_GEN_RCV_OK)
        IGNORECASE(OID_GEN_XMIT_ERROR)
        IGNORECASE(OID_GEN_RCV_ERROR)
        IGNORECASE(OID_GEN_RCV_NO_BUFFER)
        IGNORECASE(OID_GEN_DIRECTED_BYTES_XMIT)
        IGNORECASE(OID_GEN_DIRECTED_FRAMES_XMIT)
        IGNORECASE(OID_GEN_MULTICAST_BYTES_XMIT)
        IGNORECASE(OID_GEN_MULTICAST_FRAMES_XMIT)
        IGNORECASE(OID_GEN_BROADCAST_BYTES_XMIT)
        IGNORECASE(OID_GEN_BROADCAST_FRAMES_XMIT)
        IGNORECASE(OID_GEN_DIRECTED_BYTES_RCV)
        IGNORECASE(OID_GEN_DIRECTED_FRAMES_RCV)
        IGNORECASE(OID_GEN_MULTICAST_BYTES_RCV)
        IGNORECASE(OID_GEN_MULTICAST_FRAMES_RCV)
        IGNORECASE(OID_GEN_BROADCAST_BYTES_RCV)
        IGNORECASE(OID_GEN_BROADCAST_FRAMES_RCV)
        IGNORECASE(OID_GEN_RCV_CRC_ERROR)
        MAKECASE(OID_GEN_TRANSMIT_QUEUE_LENGTH)

        /* Statistical OIDs for NDIS 6.0 */
        IGNORECASE(OID_GEN_STATISTICS)
        IGNORECASE(OID_GEN_BYTES_RCV)
        IGNORECASE(OID_GEN_BYTES_XMIT)
        IGNORECASE(OID_GEN_RCV_DISCARDS)
        IGNORECASE(OID_GEN_XMIT_DISCARDS)

        /* Misc OIDs */
        MAKECASE(OID_GEN_GET_TIME_CAPS)
        MAKECASE(OID_GEN_GET_NETCARD_TIME)
        MAKECASE(OID_GEN_NETCARD_LOAD)
        MAKECASE(OID_GEN_DEVICE_PROFILE)
        MAKECASE(OID_GEN_INIT_TIME_MS)
        MAKECASE(OID_GEN_RESET_COUNTS)
        MAKECASE(OID_GEN_MEDIA_SENSE_COUNTS)

        /* PnP power management operational OIDs */
        MAKECASE(OID_PNP_CAPABILITIES)
        MAKECASE(OID_PNP_SET_POWER)
        MAKECASE(OID_PNP_QUERY_POWER)
        MAKECASE(OID_PNP_ADD_WAKE_UP_PATTERN)
        MAKECASE(OID_PNP_REMOVE_WAKE_UP_PATTERN)
        MAKECASE(OID_PNP_WAKE_UP_PATTERN_LIST)
        MAKECASE(OID_PNP_ENABLE_WAKE_UP)

        /* PnP power management statistical OIDs */
        MAKECASE(OID_PNP_WAKE_UP_OK)
        MAKECASE(OID_PNP_WAKE_UP_ERROR)

        /* Ethernet operational OIDs */
        MAKECASE(OID_802_3_PERMANENT_ADDRESS)
        MAKECASE(OID_802_3_CURRENT_ADDRESS)
        MAKECASE(OID_802_3_MULTICAST_LIST)
        MAKECASE(OID_802_3_MAXIMUM_LIST_SIZE)
        MAKECASE(OID_802_3_MAC_OPTIONS)

        /* Ethernet operational OIDs for NDIS 6.0 */
        MAKECASE(OID_802_3_ADD_MULTICAST_ADDRESS)
        MAKECASE(OID_802_3_DELETE_MULTICAST_ADDRESS)

        /* Ethernet statistical OIDs */
        MAKECASE(OID_802_3_RCV_ERROR_ALIGNMENT)
        MAKECASE(OID_802_3_XMIT_ONE_COLLISION)
        MAKECASE(OID_802_3_XMIT_MORE_COLLISIONS)
        MAKECASE(OID_802_3_XMIT_DEFERRED)
        MAKECASE(OID_802_3_XMIT_MAX_COLLISIONS)
        MAKECASE(OID_802_3_RCV_OVERRUN)
        MAKECASE(OID_802_3_XMIT_UNDERRUN)
        MAKECASE(OID_802_3_XMIT_HEARTBEAT_FAILURE)
        MAKECASE(OID_802_3_XMIT_TIMES_CRS_LOST)
        MAKECASE(OID_802_3_XMIT_LATE_COLLISIONS)

        /*  TCP/IP OIDs */
        MAKECASE(OID_TCP_TASK_OFFLOAD)
        MAKECASE(OID_TCP_TASK_IPSEC_ADD_SA)
        MAKECASE(OID_TCP_TASK_IPSEC_DELETE_SA)
        MAKECASE(OID_TCP_SAN_SUPPORT)
        MAKECASE(OID_TCP_TASK_IPSEC_ADD_UDPESP_SA)
        MAKECASE(OID_TCP_TASK_IPSEC_DELETE_UDPESP_SA)
        MAKECASE(OID_TCP4_OFFLOAD_STATS)
        MAKECASE(OID_TCP6_OFFLOAD_STATS)
        IGNORECASE(OID_IP4_OFFLOAD_STATS)
        IGNORECASE(OID_IP6_OFFLOAD_STATS)

        /* TCP offload OIDs for NDIS 6 */
        MAKECASE(OID_TCP_OFFLOAD_CURRENT_CONFIG)
        MAKECASE(OID_TCP_OFFLOAD_PARAMETERS)
        MAKECASE(OID_TCP_OFFLOAD_HARDWARE_CAPABILITIES)
        MAKECASE(OID_TCP_CONNECTION_OFFLOAD_CURRENT_CONFIG)
        MAKECASE(OID_TCP_CONNECTION_OFFLOAD_HARDWARE_CAPABILITIES)
        MAKECASE(OID_OFFLOAD_ENCAPSULATION)

#if defined(NDIS620_MINIPORT)
        MAKECASE(OID_PM_CURRENT_CAPABILITIES)
        MAKECASE(OID_PM_HARDWARE_CAPABILITIES)
        MAKECASE(OID_PM_PARAMETERS)
        MAKECASE(OID_PM_ADD_WOL_PATTERN)
        MAKECASE(OID_PM_REMOVE_WOL_PATTERN)
        MAKECASE(OID_PM_WOL_PATTERN_LIST)

        /* VMQ OIDs for NDIS 6.20 */
        MAKECASE(OID_RECEIVE_FILTER_FREE_QUEUE)
        MAKECASE(OID_RECEIVE_FILTER_CLEAR_FILTER)
        MAKECASE(OID_RECEIVE_FILTER_ALLOCATE_QUEUE)
        MAKECASE(OID_RECEIVE_FILTER_QUEUE_ALLOCATION_COMPLETE)
        MAKECASE(OID_RECEIVE_FILTER_SET_FILTER)
#endif
    }

    if (oidName)
    {
        DEBUGP(MP_LOUD, ("%s %s", Method, oidName));
    }
    else
    {
        if (!ignore)
            DEBUGP(MP_LOUD, ("<** Unknown %s OID 0x%08x **>\n", Method, Oid));
    }
}  // DbgPrintOidName
#endif


#if 0
static
NDIS_STATUS QueryTaskOffload (
    IN  PNDIS_ADAPTER             pAdapter,
    IN  UINT                      ulBufferLength,
    OUT PNDIS_TASK_OFFLOAD_HEADER pOffLoadHeader,
    OUT PUINT                     pulBytesNeeded,
    OUT PUINT                     pulBytesWritten )
{
    PNDIS_TASK_OFFLOAD         pTaskOffLoad;
    PNDIS_TASK_TCP_IP_CHECKSUM pCheckSum;
#if 0
    PNDIS_TASK_IPSEC           pIpSec;
#endif
    UINT                       uiMoveBytes;

    uiMoveBytes = sizeof( NDIS_TASK_OFFLOAD_HEADER ) +
        sizeof( NDIS_TASK_OFFLOAD ) +
        sizeof( NDIS_TASK_TCP_IP_CHECKSUM );
#if 0
    uiMoveBytes +=
        sizeof( NDIS_TASK_OFFLOAD ) +
        sizeof( NDIS_TASK_IPSEC );
#endif
    if ( ulBufferLength < uiMoveBytes )
    {
        *pulBytesNeeded = uiMoveBytes;
        return NDIS_STATUS_INVALID_LENGTH;
    }

    /* NDIS_STATUS_NOT_SUPPORTED should be returned if parameters are not
       supported by the driver.
    */
    if ( NDIS_TASK_OFFLOAD_VERSION != pOffLoadHeader->Version  ||
            sizeof( NDIS_TASK_OFFLOAD_HEADER ) != pOffLoadHeader->Size  ||
            IEEE_802_3_Encapsulation !=
            pOffLoadHeader->EncapsulationFormat.Encapsulation )
    {
        return NDIS_STATUS_NOT_SUPPORTED;
    }

    /* Tell upper edge protocol driver that task-offloading is supported. */
    pOffLoadHeader->OffsetFirstTask = sizeof( NDIS_TASK_OFFLOAD_HEADER );

    pOffLoadHeader++;
    pTaskOffLoad = ( NDIS_TASK_OFFLOAD* ) pOffLoadHeader;
    pTaskOffLoad->Version = NDIS_TASK_OFFLOAD_VERSION;
    pTaskOffLoad->Size = sizeof( NDIS_TASK_OFFLOAD );
    pTaskOffLoad->Task = TcpIpChecksumNdisTask;

    /* Currently supports only one task-offload. */
    pTaskOffLoad->OffsetNextTask = 0;
    pTaskOffLoad->TaskBufferLength = sizeof( NDIS_TASK_TCP_IP_CHECKSUM );

    /* Fill checksum offload capabilities. */
    pCheckSum = ( NDIS_TASK_TCP_IP_CHECKSUM* ) pTaskOffLoad->TaskBuffer;

    if (pAdapter->ulTxChecksum)
    {
        pCheckSum->V4Transmit.IpOptionsSupported = FALSE;
        pCheckSum->V4Transmit.TcpOptionsSupported = TRUE;

        pCheckSum->V4Transmit.TcpChecksum = TRUE;
        pCheckSum->V4Transmit.UdpChecksum = TRUE;
        pCheckSum->V4Transmit.IpChecksum = TRUE;
    }
    else
    {
        pCheckSum->V4Transmit.IpOptionsSupported = FALSE;
        pCheckSum->V4Transmit.TcpOptionsSupported = FALSE;
        pCheckSum->V4Transmit.TcpChecksum = FALSE;
        pCheckSum->V4Transmit.UdpChecksum = FALSE;
        pCheckSum->V4Transmit.IpChecksum = FALSE;
    }

    if (pAdapter->ulRxChecksum)
    {
        pCheckSum->V4Receive.IpOptionsSupported = TRUE;
        pCheckSum->V4Receive.TcpOptionsSupported = TRUE;

        pCheckSum->V4Receive.TcpChecksum = TRUE;
        pCheckSum->V4Receive.UdpChecksum = FALSE;
        pCheckSum->V4Receive.IpChecksum = TRUE;
    }
    else
    {
        pCheckSum->V4Receive.IpOptionsSupported = FALSE;
        pCheckSum->V4Receive.TcpOptionsSupported = FALSE;
        pCheckSum->V4Receive.TcpChecksum = FALSE;
        pCheckSum->V4Receive.UdpChecksum = FALSE;
        pCheckSum->V4Receive.IpChecksum = FALSE;
    }

    /* V6 is not supported for now. */
    pCheckSum->V6Transmit.IpOptionsSupported = FALSE;
    pCheckSum->V6Transmit.TcpOptionsSupported = FALSE;
    pCheckSum->V6Transmit.TcpChecksum = FALSE;
    pCheckSum->V6Transmit.UdpChecksum = FALSE;
    pCheckSum->V6Receive.IpOptionsSupported = FALSE;
    pCheckSum->V6Receive.TcpOptionsSupported = FALSE;
    pCheckSum->V6Receive.TcpChecksum = FALSE;
    pCheckSum->V6Receive.UdpChecksum = FALSE;

#if 0
    pTaskOffLoad->OffsetNextTask =
        sizeof( NDIS_TASK_OFFLOAD ) +
        sizeof( NDIS_TASK_TCP_IP_CHECKSUM );

    pTaskOffLoad = ( NDIS_TASK_OFFLOAD* )(( PUCHAR ) pTaskOffLoad +
        pTaskOffLoad->OffsetNextTask );
    pTaskOffLoad->Version = NDIS_TASK_OFFLOAD_VERSION;
    pTaskOffLoad->Size = sizeof( NDIS_TASK_OFFLOAD );
    pTaskOffLoad->Task = IpSecNdisTask;

    /* Currently supports only one task-offload. */
    pTaskOffLoad->OffsetNextTask = 0;
    pTaskOffLoad->TaskBufferLength = sizeof( NDIS_TASK_IPSEC );

    pIpSec = ( NDIS_TASK_IPSEC* ) pTaskOffLoad->TaskBuffer;

    pIpSec->Supported.AH_ESP_COMBINED = TRUE;
    pIpSec->Supported.TRANSPORT_TUNNEL_COMBINED = TRUE;
    pIpSec->Supported.V4_OPTIONS = TRUE;
    pIpSec->Supported.RESERVED = 0;

    pIpSec->V4AH.MD5 = TRUE;
    pIpSec->V4AH.SHA_1 = TRUE;
    pIpSec->V4AH.Transport = TRUE;
    pIpSec->V4AH.Tunnel = TRUE;
    pIpSec->V4AH.Send = TRUE;
    pIpSec->V4AH.Receive = TRUE;

    pIpSec->V4ESP.DES = TRUE;
    pIpSec->V4ESP.TRIPLE_DES = TRUE;
    pIpSec->V4ESP.NULL_ESP = FALSE;
    pIpSec->V4ESP.Transport = TRUE;
    pIpSec->V4ESP.Tunnel = TRUE;
    pIpSec->V4ESP.Send = TRUE;
    pIpSec->V4ESP.Receive = TRUE;
#endif

    *pulBytesWritten += uiMoveBytes;

#if DBG
DBG_PRINT( "query offload: %u"NEWLINE, *pulBytesWritten );
#endif
    return NDIS_STATUS_SUCCESS;
}  /* QueryTaskOffload */


static
NDIS_STATUS SetTaskOffload (
    IN  PNDIS_ADAPTER             pAdapter,
    IN  PNDIS_TASK_OFFLOAD_HEADER pOffLoadHeader,
    OUT PUINT                     pulBytesRead )
{
    PNDIS_TASK_OFFLOAD         pTaskOffLoad;
    PNDIS_TASK_TCP_IP_CHECKSUM pCheckSum;
    PNDIS_TASK_IPSEC           pIpSec;
    BOOLEAN                    bFound;
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;

    if ( sizeof( NDIS_TASK_OFFLOAD_HEADER ) != pOffLoadHeader->Size  ||
            NDIS_TASK_OFFLOAD_VERSION != pOffLoadHeader->Version  ||
            IEEE_802_3_Encapsulation !=
            pOffLoadHeader->EncapsulationFormat.Encapsulation )
    {
        return NDIS_STATUS_NOT_SUPPORTED;
    }
    *pulBytesRead += sizeof( NDIS_TASK_OFFLOAD_HEADER );

    /* All task-offloading are disabled. */
    if ( 0 == pOffLoadHeader->OffsetFirstTask )
    {
        if (pAdapter->ulTxChecksum)
        {
            pAdapter->ulTxChecksum &= ~(
                TASK_OFFLOAD_IP_CHECKSUM |
                TASK_OFFLOAD_TCP_CHECKSUM |
                TASK_OFFLOAD_UDP_CHECKSUM );
        }
        if (pAdapter->ulRxChecksum)
        {
            pAdapter->ulRxChecksum &= ~(
                TASK_OFFLOAD_IP_CHECKSUM |
                TASK_OFFLOAD_TCP_CHECKSUM |
                TASK_OFFLOAD_UDP_CHECKSUM );
        }
        return Status;
    }

    /* Look through task chain to see whether suitable task(s) were set by the
       upper edge driver and take the last one to set.
    */
    pTaskOffLoad = ( NDIS_TASK_OFFLOAD* )(( PUCHAR ) pOffLoadHeader +
        pOffLoadHeader->OffsetFirstTask );
    bFound = FALSE;
    while ( pTaskOffLoad )
    {
        *pulBytesRead += sizeof( NDIS_TASK_OFFLOAD );

        /* Check to see whether it is the checksum task we support. */
        if ( TcpIpChecksumNdisTask == pTaskOffLoad->Task )
        {
            pCheckSum = ( NDIS_TASK_TCP_IP_CHECKSUM* )
                pTaskOffLoad->TaskBuffer;
            *pulBytesRead += sizeof( NDIS_TASK_TCP_IP_CHECKSUM );

            /* We need to detect to see whether unsupported settings are passed
               from upper edge before following the instruction.
            */
            if ( pCheckSum->V4Transmit.TcpOptionsSupported  ||
                    pCheckSum->V4Transmit.IpOptionsSupported )
            {
                pCheckSum->V4Transmit.TcpOptionsSupported = FALSE;
                pCheckSum->V4Transmit.IpOptionsSupported = FALSE;
                return NDIS_STATUS_NOT_SUPPORTED;
            }
            if ( pCheckSum->V4Receive.TcpOptionsSupported  ||
                    pCheckSum->V4Receive.IpOptionsSupported )
            {
                pCheckSum->V4Receive.TcpOptionsSupported = FALSE;
                pCheckSum->V4Receive.IpOptionsSupported = FALSE;
                return NDIS_STATUS_NOT_SUPPORTED;
            }
            bFound = TRUE;
            if (pAdapter->ulTxChecksum)
            {
                if ( pCheckSum->V4Transmit.TcpChecksum )
                    pAdapter->ulTxChecksum |= TASK_OFFLOAD_TCP_CHECKSUM;
                if ( pCheckSum->V4Transmit.UdpChecksum )
                    pAdapter->ulTxChecksum |= TASK_OFFLOAD_UDP_CHECKSUM;
                if ( pCheckSum->V4Transmit.IpChecksum )
                    pAdapter->ulTxChecksum |= TASK_OFFLOAD_IP_CHECKSUM;
            }
            if (pAdapter->ulRxChecksum)
            {
                PHARDWARE pHardware = &pAdapter->Hardware;

                if ( pCheckSum->V4Receive.TcpChecksum )
                {
                    pAdapter->ulRxChecksum |= TASK_OFFLOAD_TCP_CHECKSUM;
                    pHardware->m_dwReceiveConfig |=
                        DMA_RX_CTRL_CSUM_TCP;
                }
                if ( pCheckSum->V4Receive.UdpChecksum )
                {
                    pAdapter->ulRxChecksum |= TASK_OFFLOAD_UDP_CHECKSUM;
                    pHardware->m_dwReceiveConfig |=
                        DMA_RX_CTRL_CSUM_UDP;
                }
                if ( pCheckSum->V4Receive.IpChecksum )
                {
                    pAdapter->ulRxChecksum |= TASK_OFFLOAD_IP_CHECKSUM;
                    pHardware->m_dwReceiveConfig |=
                        DMA_RX_CTRL_CSUM_IP;
                }
                HardwareStopReceive( pHardware );
                if ( pHardware->m_bEnabled )
                    HardwareStartReceive( pHardware );
            }

#if DBG
DBG_PRINT( "checksum offload: %x %x; %u"NEWLINE, pAdapter->ulRxChecksum,
    pAdapter->ulTxChecksum, *pulBytesRead );
#endif
        }
        else if ( IpSecNdisTask == pTaskOffLoad->Task )
        {
            pIpSec = ( NDIS_TASK_IPSEC* )
                pTaskOffLoad->TaskBuffer;
            *pulBytesRead += sizeof( NDIS_TASK_IPSEC );

#if DBG
DBG_PRINT( "ipsec: %d %d %d %d; %d %d %d %d %d %d; %d %d %d %d %d %d %d"NEWLINE,
    pIpSec->Supported.AH_ESP_COMBINED,
    pIpSec->Supported.TRANSPORT_TUNNEL_COMBINED,
    pIpSec->Supported.V4_OPTIONS,
    pIpSec->Supported.RESERVED,
    pIpSec->V4AH.MD5,
    pIpSec->V4AH.SHA_1,
    pIpSec->V4AH.Transport,
    pIpSec->V4AH.Tunnel,
    pIpSec->V4AH.Send,
    pIpSec->V4AH.Receive,
    pIpSec->V4ESP.DES,
    pIpSec->V4ESP.TRIPLE_DES,
    pIpSec->V4ESP.NULL_ESP,
    pIpSec->V4ESP.Transport,
    pIpSec->V4ESP.Tunnel,
    pIpSec->V4ESP.Send,
    pIpSec->V4ESP.Receive );
#endif

            if ( pIpSec->V4ESP.NULL_ESP )
            {
                pIpSec->V4ESP.NULL_ESP = FALSE;
                return NDIS_STATUS_NOT_SUPPORTED;
            }
        }

        /* To next offload task if any. */
        if ( 0 == pTaskOffLoad->OffsetNextTask )
            break;

        pTaskOffLoad = ( NDIS_TASK_OFFLOAD* )(( PUCHAR ) pTaskOffLoad +
            pTaskOffLoad->OffsetNextTask );
    }

    return( FALSE == bFound ? NDIS_STATUS_NOT_SUPPORTED : Status );
}  /* SetTaskOffload */
#endif


static
NDIS_STATUS AddWakeUpPattern (
    IN  PNDIS_ADAPTER           pAdapter,
    IN  PNDIS_PM_PACKET_PATTERN pPattern,
    OUT PUINT                   pulBytesRead )
{
    NDIS_STATUS Status;
    PWAKE_UP_PACKET pWakeUpPacket;

    if (pAdapter->cnWakeUpPacket < MAX_WAKE_UP_PACKET &&
            pPattern->MaskSize <= MAX_PATTERN_LENGTH / 8 &&
            pPattern->PatternSize <= MAX_PATTERN_LENGTH)
    {
        int nIndex;

        for (nIndex = 0; nIndex < MAX_WAKE_UP_PACKET; nIndex++)
        {
            pWakeUpPacket = &pAdapter->WakeUpPacket[nIndex];
            if (!pWakeUpPacket->NdisPacketPattern.MaskSize)
                break;
        }

        /* Move the packet pattern separately in case they are not
           continuous.
        */
        pAdapter->cnWakeUpPacket++;

        /* Copy the NDIS_PM_PACKET_PATTERN. */
        pWakeUpPacket->NdisPacketPattern = *pPattern;
        pWakeUpPacket->NdisPacketPattern.PatternOffset =
            sizeof(NDIS_PM_PACKET_PATTERN) + pPattern->MaskSize;
        *pulBytesRead += sizeof(NDIS_PM_PACKET_PATTERN);

        /* Copy the MASK. */
        NdisMoveMemory(
            &pWakeUpPacket->MaskAndPattern[0], (pPattern + 1),
            pPattern->MaskSize);
        *pulBytesRead += pPattern->MaskSize;

        /* Copy the PATTERN. */
        NdisMoveMemory(
            &pWakeUpPacket->MaskAndPattern[pPattern->MaskSize],
            (PUCHAR) pPattern + pPattern->PatternOffset,
            pPattern->PatternSize);
        *pulBytesRead += pPattern->PatternSize;

#if DBG_
{
    UINT i;

    for (i = 0; i < pPattern->MaskSize; i++)
        DBG_PRINT("%02X ", pWakeUpPacket->MaskAndPattern[i]);
    DBG_PRINT(NEWLINE);
    for (; i < pPattern->MaskSize + pPattern->PatternSize; i++)
        DBG_PRINT("%02X ", pWakeUpPacket->MaskAndPattern[i]);
    DBG_PRINT(NEWLINE);
}
DBG_PRINT("Added: %u"NEWLINE, *pulBytesRead);
#endif
        HardwareSetWolFrame(&pAdapter->Hardware, nIndex,
            pPattern->MaskSize,
            pWakeUpPacket->MaskAndPattern,
            pPattern->PatternSize,
            &pWakeUpPacket->MaskAndPattern[pPattern->MaskSize]);

        Status = NDIS_STATUS_SUCCESS;
    }
    else
    {
        *pulBytesRead = 0;
        Status = NDIS_STATUS_RESOURCES;
    }

    return Status;
}  /* AddWakeUpPattern */


static
NDIS_STATUS RemoveWakeUpPattern (
    IN  PNDIS_ADAPTER           pAdapter,
    IN  PNDIS_PM_PACKET_PATTERN pPattern,
    OUT PUINT                   pulBytesRead )
{
    NDIS_STATUS Status;
    PWAKE_UP_PACKET pWakeUpPacket;
    UINT            nCount;
    int             nIndex;
    PUCHAR          pSource;
    PUCHAR          pTarget;

    for (nIndex = 0; nIndex < MAX_WAKE_UP_PACKET; nIndex++)
    {
        pWakeUpPacket = &pAdapter->WakeUpPacket[nIndex];
        if (pPattern->MaskSize == pWakeUpPacket->NdisPacketPattern.MaskSize
                && pPattern->PatternSize ==
                pWakeUpPacket->NdisPacketPattern.PatternSize)
        {
            /* Compare the mask. */
            for (nCount = 0,
                    pSource = (PUCHAR) pPattern +
                    sizeof(NDIS_PM_PACKET_PATTERN),
                    pTarget = pWakeUpPacket->MaskAndPattern;
                    nCount < pWakeUpPacket->NdisPacketPattern.MaskSize;
                    nCount++)
                if (pSource[nCount] != pTarget[nCount])
                    break;
            if (nCount < pWakeUpPacket->NdisPacketPattern.MaskSize)
                continue;

            /* Compare the pattern. */
            for (nCount = 0,
                    pSource = (PUCHAR) pPattern + pPattern->PatternOffset,
                    pTarget = &pWakeUpPacket->MaskAndPattern[
                        pPattern->MaskSize];
                    nCount < pWakeUpPacket->NdisPacketPattern.PatternSize;
                    nCount++)
                if (pSource[nCount] != pTarget[nCount])
                    break;
            if (nCount == pWakeUpPacket->NdisPacketPattern.PatternSize)
            {
                pWakeUpPacket->NdisPacketPattern.MaskSize = 0;
                break;
            }
        }
    }

    if (nIndex < MAX_WAKE_UP_PACKET)
    {
        pAdapter->cnWakeUpPacket--;

        *pulBytesRead = sizeof(NDIS_PM_PACKET_PATTERN) +
            pPattern->MaskSize + pPattern->PatternSize;

#if DBG
DBG_PRINT("Remove: %d, %u"NEWLINE, nIndex, *pulBytesRead);
#endif
        HardwareSetWolEnable(&pAdapter->Hardware,
            (USHORT)(1 << nIndex), FALSE);
        Status = NDIS_STATUS_SUCCESS;
    }
    else
    {
        *pulBytesRead = 0;
        Status = NDIS_STATUS_NOT_ACCEPTED;
    }

    return Status;
}  /* RemoveWakeUpPattern */


static
NDIS_STATUS QueryStatistics (
    IN  PNDIS_ADAPTER pAdapter,
    OUT struct _QUERY *Query )
{
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;
    PNDIS_STATISTICS_INFO Statistics = (PNDIS_STATISTICS_INFO)
        Query->InformationBuffer;

    if (Query->InformationBufferLength < sizeof(NDIS_STATISTICS_INFO))
    {
        Status = NDIS_STATUS_INVALID_LENGTH;
        Query->BytesNeeded = sizeof(NDIS_STATISTICS_INFO);
        return Status;
    }
    Statistics->Header.Type = NDIS_OBJECT_TYPE_DEFAULT;
    Statistics->Header.Size = NDIS_SIZEOF_STATISTICS_INFO_REVISION_1;
    Statistics->Header.Revision = NDIS_STATISTICS_INFO_REVISION_1;

    Statistics->SupportedStatistics =
        NDIS_STATISTICS_FLAGS_VALID_DIRECTED_FRAMES_RCV |
        NDIS_STATISTICS_FLAGS_VALID_MULTICAST_FRAMES_RCV |
        NDIS_STATISTICS_FLAGS_VALID_BROADCAST_FRAMES_RCV |
        NDIS_STATISTICS_FLAGS_VALID_BYTES_RCV |
        NDIS_STATISTICS_FLAGS_VALID_RCV_DISCARDS |
        NDIS_STATISTICS_FLAGS_VALID_RCV_ERROR |
        NDIS_STATISTICS_FLAGS_VALID_DIRECTED_FRAMES_XMIT |
        NDIS_STATISTICS_FLAGS_VALID_MULTICAST_FRAMES_XMIT |
        NDIS_STATISTICS_FLAGS_VALID_BROADCAST_FRAMES_XMIT |
        NDIS_STATISTICS_FLAGS_VALID_BYTES_XMIT |
        NDIS_STATISTICS_FLAGS_VALID_XMIT_ERROR |
        NDIS_STATISTICS_FLAGS_VALID_XMIT_DISCARDS |
        NDIS_STATISTICS_FLAGS_VALID_DIRECTED_BYTES_RCV |
        NDIS_STATISTICS_FLAGS_VALID_MULTICAST_BYTES_RCV |
        NDIS_STATISTICS_FLAGS_VALID_BROADCAST_BYTES_RCV |
        NDIS_STATISTICS_FLAGS_VALID_DIRECTED_BYTES_XMIT |
        NDIS_STATISTICS_FLAGS_VALID_MULTICAST_BYTES_XMIT |
        NDIS_STATISTICS_FLAGS_VALID_BROADCAST_BYTES_XMIT;

    Statistics->ifInDiscards =
        AdapterReadCounter(pAdapter, OID_GEN_RCV_DISCARDS);
    Statistics->ifInErrors =
        AdapterReadCounter(pAdapter, OID_GEN_RCV_ERROR);
    Statistics->ifHCInOctets =
        AdapterReadCounter(pAdapter, OID_GEN_BYTES_RCV);
    Statistics->ifHCInUcastPkts =
        AdapterReadCounter(pAdapter, OID_GEN_DIRECTED_FRAMES_RCV);
    Statistics->ifHCInMulticastPkts =
        AdapterReadCounter(pAdapter, OID_GEN_MULTICAST_FRAMES_RCV);
    Statistics->ifHCInBroadcastPkts =
        AdapterReadCounter(pAdapter, OID_GEN_BROADCAST_FRAMES_RCV);
    Statistics->ifHCOutOctets =
        AdapterReadCounter(pAdapter, OID_GEN_BYTES_XMIT);
    Statistics->ifHCOutUcastPkts =
        AdapterReadCounter(pAdapter, OID_GEN_DIRECTED_FRAMES_XMIT);
    Statistics->ifHCOutMulticastPkts =
        AdapterReadCounter(pAdapter, OID_GEN_MULTICAST_FRAMES_XMIT);
    Statistics->ifHCOutBroadcastPkts =
        AdapterReadCounter(pAdapter, OID_GEN_BROADCAST_FRAMES_XMIT);
    Statistics->ifOutErrors =
        AdapterReadCounter(pAdapter, OID_GEN_XMIT_ERROR);
    Statistics->ifOutDiscards =
        AdapterReadCounter(pAdapter, OID_GEN_XMIT_DISCARDS);
    Statistics->ifHCInUcastOctets =
        AdapterReadCounter(pAdapter, OID_GEN_DIRECTED_BYTES_RCV);
    Statistics->ifHCInMulticastOctets =
        AdapterReadCounter(pAdapter, OID_GEN_MULTICAST_BYTES_RCV);
    Statistics->ifHCInBroadcastOctets =
        AdapterReadCounter(pAdapter, OID_GEN_BROADCAST_BYTES_RCV);
    Statistics->ifHCOutUcastOctets =
        AdapterReadCounter(pAdapter, OID_GEN_DIRECTED_BYTES_XMIT);
    Statistics->ifHCOutMulticastOctets =
        AdapterReadCounter(pAdapter, OID_GEN_MULTICAST_BYTES_XMIT);
    Statistics->ifHCOutBroadcastOctets =
        AdapterReadCounter(pAdapter, OID_GEN_BROADCAST_BYTES_XMIT);

    Query->BytesWritten = NDIS_SIZEOF_STATISTICS_INFO_REVISION_1;
    return Status;
}  /* QueryStatistics */

/* -------------------------------------------------------------------------- */

/*
    MiniportQueryInformation

Description:
    This function is used by NDIS to query adapter information by
    pre-defined NDIS_OIDs.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PNDIS_OID_Request NdisQueryRequest
        The NDIS_OID request to process.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

NDIS_STATUS MiniportQueryInformation (
    IN  NDIS_HANDLE       hAdapterContext,
    OUT PNDIS_OID_REQUEST NdisQueryRequest )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;
    struct _QUERY *Query = &NdisQueryRequest->DATA.QUERY_INFORMATION;
    PHARDWARE pHardware = &pAdapter->Hardware;
    PPORT_CONFIG pPort = &pHardware->m_Port[MAIN_PORT];
    ULONGLONG qCounter;
    ULONG ulInfo;
    USHORT usInfo;
    ULONG64 ulInfo64;

    // Default to returning the ULONG value
    PVOID pInfo = &ulInfo;
    ULONG ulInfoLen = sizeof(ulInfo);

#ifdef DBG
    DbgPrintOidName("query", Query->Oid);
#endif

    Query->BytesWritten = 0;
    Query->BytesNeeded = 0;
    switch (Query->Oid)
    {
        case OID_GEN_HARDWARE_STATUS:
        {
            NDIS_HARDWARE_STATUS HardwareStatus = NdisHardwareStatusReady;

            //
            // Specify the current hardware status of the underlying NIC as
            // one of the following NDIS_HARDWARE_STATUS-type values.
            //
#if 0
            HardwareStatus = NdisHardwareStatusNotReady;
            HardwareStatus = NdisHardwareStatusReset;
#endif
            pInfo = (PVOID) &HardwareStatus;
            ulInfoLen = sizeof(NDIS_HARDWARE_STATUS);
            break;
        }

        case OID_GEN_MAXIMUM_TOTAL_SIZE:
            //
            // Specify the maximum total packet length, in bytes, the NIC
            // supports including the header. A protocol driver might use
            // this returned length as a gauge to determine the maximum
            // size packet that a NIC driver could forward to the
            // protocol driver. The miniport driver must never indicate
            // up to the bound protocol driver packets received over the
            // network that are longer than the packet size specified by
            // OID_GEN_MAXIMUM_TOTAL_SIZE.
            //

            __fallthrough;

        case OID_GEN_TRANSMIT_BLOCK_SIZE:
            //
            // The OID_GEN_TRANSMIT_BLOCK_SIZE OID specifies the minimum
            // number of bytes that a single net packet occupies in the
            // transmit buffer space of the NIC. For example, a NIC that
            // has a transmit space divided into 256-byte pieces would have
            // a transmit block size of 256 bytes. To calculate the total
            // transmit buffer space on such a NIC, its driver multiplies
            // the number of transmit buffers on the NIC by its transmit
            // block size. In our case, the transmit block size is
            // identical to its maximum packet size.

            __fallthrough;

        case OID_GEN_RECEIVE_BLOCK_SIZE:
            //
            // The OID_GEN_RECEIVE_BLOCK_SIZE OID specifies the amount of
            // storage, in bytes, that a single packet occupies in the receive
            // buffer space of the NIC.
            //
            ulInfo = (ULONG) MAXIMUM_ETHERNET_PACKET_SIZE;
            break;

        case OID_GEN_TRANSMIT_BUFFER_SPACE:
            //
            // Specify the amount of memory, in bytes, on the NIC that
            // is available for buffering transmit data. A protocol can
            // use this OID as a guide for sizing the amount of transmit
            // data per send.
            //
#if 0
            ulInfo = HW_MAX_FRAME_SIZE * Adapter->ulMaxBusySends;
#endif
            ulInfo = TX_BUF_SIZE;
            break;

        case OID_GEN_RECEIVE_BUFFER_SPACE:
            //
            // Specify the amount of memory on the NIC that is available
            // for buffering receive data. A protocol driver can use this
            // OID as a guide for advertising its receive window after it
            // establishes sessions with remote nodes.
            //
#if 0
            ulInfo = HW_MAX_FRAME_SIZE * Adapter->ulMaxBusyRecvs;
#endif
            ulInfo = RX_BUF_SIZE;
            break;

        case OID_GEN_MEDIA_SUPPORTED:
            //
            // Return an array of media that are supported by the miniport.
            // This miniport only supports one medium (Ethernet), so the OID
            // returns identical results to OID_GEN_MEDIA_IN_USE.
            //

            __fallthrough;

        case OID_GEN_MEDIA_IN_USE:
        {
            NDIS_MEDIUM Medium = NdisMedium802_3;
            //
            // Return an array of media that are currently in use by the
            // miniport.  This array should be a subset of the array returned
            // by OID_GEN_MEDIA_SUPPORTED.
            //
            pInfo = &Medium;
            ulInfoLen = sizeof(Medium);
            break;
        }

        case OID_GEN_MAXIMUM_SEND_PACKETS:
            ulInfo = pAdapter->TxBufInfo.cnAlloc >> 1;

#ifdef DBG
            DBG_PRINT("max send packets: %d"NEWLINE, ulInfo);
#endif
            break;

        // counters
        case OID_GEN_XMIT_OK:
        case OID_GEN_RCV_OK:
        case OID_GEN_XMIT_ERROR:
        case OID_GEN_RCV_ERROR:
        case OID_GEN_RCV_NO_BUFFER:
        case OID_GEN_DIRECTED_BYTES_XMIT:
        case OID_GEN_DIRECTED_FRAMES_XMIT:
        case OID_GEN_BROADCAST_BYTES_XMIT:
        case OID_GEN_BROADCAST_FRAMES_XMIT:
        case OID_GEN_MULTICAST_BYTES_XMIT:
        case OID_GEN_MULTICAST_FRAMES_XMIT:
        case OID_GEN_DIRECTED_BYTES_RCV:
        case OID_GEN_DIRECTED_FRAMES_RCV:
        case OID_GEN_BROADCAST_BYTES_RCV:
        case OID_GEN_BROADCAST_FRAMES_RCV:
        case OID_GEN_MULTICAST_BYTES_RCV:
        case OID_GEN_MULTICAST_FRAMES_RCV:
        case OID_GEN_RCV_CRC_ERROR:
        case OID_802_3_RCV_ERROR_ALIGNMENT:
        case OID_802_3_XMIT_ONE_COLLISION:
        case OID_802_3_XMIT_MORE_COLLISIONS:
        case OID_802_3_XMIT_DEFERRED:
        case OID_802_3_XMIT_MAX_COLLISIONS:
        case OID_802_3_XMIT_LATE_COLLISIONS:
        case OID_GEN_BYTES_RCV:
        case OID_GEN_BYTES_XMIT:
        case OID_GEN_RCV_DISCARDS:
        case OID_GEN_XMIT_DISCARDS:
            ulInfo64 = AdapterReadCounter(pAdapter, Query->Oid);
            pInfo = &ulInfo64;
            break;

        case OID_GEN_VENDOR_ID:
            //
            // Specify a three-byte IEEE-registered vendor code, followed
            // by a single byte that the vendor assigns to identify a
            // particular NIC. The IEEE code uniquely identifies the vendor
            // and is the same as the three bytes appearing at the beginning
            // of the NIC hardware address. Vendors without an IEEE-registered
            // code should use the value 0xFFFFFF.
            //
#ifdef DBG
            DBG_PRINT("vendor id"NEWLINE);
#endif
            NdisMoveMemory(&ulInfo,
                pAdapter->Hardware.m_bPermanentAddress, 3);
            ulInfo &= 0x00FFFFFF;
            break;

        case OID_GEN_VENDOR_DESCRIPTION:
            //
            // Specify a zero-terminated string describing the NIC vendor.
            //
#ifdef DBG
            DBG_PRINT("vendor descrip"NEWLINE);
#endif
            AdapterGetDescription(&pInfo, &ulInfoLen);
            break;

        case OID_GEN_VENDOR_DRIVER_VERSION:
            //
            // Specify the vendor-assigned version number of the NIC driver.
            // The low-order half of the return value specifies the minor
            // version; the high-order half specifies the major version.
            //
            ulInfo = ((USHORT) DRIVER_MAJOR_VERSION << 16) |
                (USHORT) DRIVER_MINOR_VERSION;
            break;

        case OID_GEN_DRIVER_VERSION:
            //
            // Specify the NDIS version in use by the NIC driver. The high
            // byte is the major version number; the low byte is the minor
            // version number.
            //
            usInfo = ((USHORT) NDIS_MINIPORT_MAJOR_VERSION << 8) |
                (USHORT) NDIS_MINIPORT_MINOR_VERSION;
            pInfo = (PVOID) &usInfo;
            ulInfoLen = sizeof(USHORT);
            break;

#if !defined(NDIS620_MINIPORT)
        case OID_PNP_CAPABILITIES:
            //
            // Return the wake-up capabilities of its NIC. If you return
            // NDIS_STATUS_NOT_SUPPORTED, NDIS considers the miniport driver
            // to be not Power management aware and doesn't send any power
            // or wake-up related queries such as
            // OID_PNP_SET_POWER, OID_PNP_QUERY_POWER,
            // OID_PNP_ADD_WAKE_UP_PATTERN, OID_PNP_REMOVE_WAKE_UP_PATTERN,
            // OID_PNP_ENABLE_WAKE_UP.
            //
            ulInfoLen = sizeof(NDIS_PNP_CAPABILITIES);
            pInfo = (PVOID) &pAdapter->PnpCapabilities;
            break;
#endif
        case OID_PNP_QUERY_POWER:
        {
            NDIS_DEVICE_POWER_STATE PowerState =
                *((PNDIS_DEVICE_POWER_STATE) Query->InformationBuffer);

#if DBG
DBG_PRINT( "query power: %d"NEWLINE, PowerState);
#endif
            switch (PowerState)
            {
                case NdisDeviceStateUnspecified:
                    break;
                case NdisDeviceStateD0:
                    break;
                case NdisDeviceStateD1:
                    break;
                case NdisDeviceStateD2:
                    break;
                case NdisDeviceStateD3:
                    break;
            }
            ulInfoLen = 0;
            break;
        }
        case OID_PNP_ENABLE_WAKE_UP:
            ulInfo = pAdapter->ulWakeUpEnable;
            break;
        case OID_PNP_WAKE_UP_OK:
            ulInfo = pAdapter->ulWakeUpOk;
            break;
        case OID_PNP_WAKE_UP_ERROR:
            ulInfo = pAdapter->ulWakeUpError;
            break;

            //
            // Following 4 OIDs are for querying Ethernet Operational
            // Characteristics.
            //

        case OID_802_3_PERMANENT_ADDRESS:
            //
            // Return the MAC address of the NIC burnt in the hardware.
            //
            pInfo = pAdapter->Hardware.m_bPermanentAddress;
            ulInfoLen = MAC_ADDRESS_LENGTH;
            break;
        case OID_802_3_CURRENT_ADDRESS:
            //
            // Return the MAC address the NIC is currently programmed to
            // use. Note that this address could be different from the
            // permananent address as the user can override using
            // registry. Read NdisReadNetworkAddress doc for more info.
            //
            pInfo = pAdapter->Hardware.m_bOverrideAddress;
            ulInfoLen = MAC_ADDRESS_LENGTH;
            break;

        case OID_802_3_MAXIMUM_LIST_SIZE:
            //
            // The maximum number of multicast addresses the NIC driver
            // can manage. This list is global for all protocols bound
            // to (or above) the NIC. Consequently, a protocol can receive
            // NDIS_STATUS_MULTICAST_FULL from the NIC driver when
            // attempting to set the multicast address list, even if
            // the number of elements in the given list is less than
            // the number originally returned for this query.
            //
            ulInfo = MAX_MULTICAST_LIST;
            break;
        case OID_802_3_MULTICAST_LIST:
            ulInfo = pAdapter->Hardware.m_bMulticastListSize;
            break;

            //
            // Following list consists of both general and Ethernet
            // specific statistical OIDs.
            //

        case OID_GEN_STATISTICS:
            return QueryStatistics(pAdapter, Query);

        case OID_GEN_TRANSMIT_QUEUE_LENGTH:
            ulInfo = pAdapter->nBusySend;
#ifdef DBG
            DBG_PRINT("transmit queue len: %u"NEWLINE, ulInfo);
#endif
            break;

        case OID_GEN_MEDIA_CONNECT_STATUS:
            ulInfo = pAdapter->ulNdisMediaState;
            break;

        case OID_GEN_PHYSICAL_MEDIUM:
            ulInfo = NdisPhysicalMediumUnspecified;
            break;

        case OID_GEN_SUPPORTED_LIST:
            pInfo = (PVOID) SupportedOids;
            ulInfoLen = pAdapter->OidsSize;
            break;

        case OID_GEN_MAXIMUM_LOOKAHEAD:
            ulInfo = NDIS_MAX_LOOKAHEAD;
#ifdef DBG
            DBG_PRINT("max lookahead: %u"NEWLINE, ulInfo);
#endif
            break;

        case OID_GEN_MAXIMUM_FRAME_SIZE:
            ulInfo = MAX_ETHERNET_BODY_SIZE;
            break;

        case OID_GEN_LINK_SPEED:
            ulInfo = pAdapter->Hardware.m_ulTransmitRate;
            break;

        case OID_GEN_CURRENT_PACKET_FILTER:
            ulInfo = pAdapter->ulPacketFilter;
            break;

        case OID_GEN_CURRENT_LOOKAHEAD:
#ifdef DBG
            DBG_PRINT("get max lookahead: %u"NEWLINE,
                pAdapter->ulMaxLookAhead);
#endif
            ulInfo = pAdapter->ulMaxLookAhead;
            break;

        case OID_GEN_MAC_OPTIONS:
            ulInfo = pAdapter->ulMacOptions;
            break;

        case OID_GEN_SUPPORTED_GUIDS:
            ulInfoLen = 0;
            break;

        case OID_IP4_OFFLOAD_STATS:
        case OID_IP6_OFFLOAD_STATS:
            Status = NDIS_STATUS_NOT_SUPPORTED;
            break;

#if 0
        case OID_TCP_TASK_OFFLOAD:
            return QueryTaskOffload(pAdapter, ulBufferLength,
                (PNDIS_TASK_OFFLOAD_HEADER) pBuffer, pulBytesNeeded,
                pulBytesWritten);

#if 0
        case OID_TCP_TASK_IPSEC_ADD_SA:
            nsStatus = NDIS_STATUS_FAILURE;
            break;

        case OID_TCP_TASK_IPSEC_DELETE_SA:
            nsStatus = NDIS_STATUS_FAILURE;
            break;
#endif

        case OID_FFP_SUPPORT:

#if DBG
            DBG_PRINT("Fast Forward Packet"NEWLINE);
#endif
            nsStatus = NDIS_STATUS_NOT_SUPPORTED;
            break;
#endif

        case OID_DEVICE_MODE:
            Status = GetDeviceMode(pAdapter, Query->Oid,
                Query->InformationBuffer, Query->InformationBufferLength,
                &Query->BytesWritten, &Query->BytesNeeded);
            ulInfoLen = 0;
            break;

        case OID_GEN_INTERRUPT_MODERATION:
        {
            PNDIS_INTERRUPT_MODERATION_PARAMETERS Moderation =
                (PNDIS_INTERRUPT_MODERATION_PARAMETERS)
                Query->InformationBuffer;

            Moderation->Header.Type = NDIS_OBJECT_TYPE_DEFAULT;
            Moderation->Header.Revision =
                NDIS_INTERRUPT_MODERATION_PARAMETERS_REVISION_1;
            Moderation->Header.Size =
                NDIS_SIZEOF_INTERRUPT_MODERATION_PARAMETERS_REVISION_1;
            Moderation->Flags = 0;
            Moderation->InterruptModeration =
                NdisInterruptModerationNotSupported;
            ulInfoLen = NDIS_SIZEOF_INTERRUPT_MODERATION_PARAMETERS_REVISION_1;
            break;
        }

        default:
#ifdef DBG
            DBG_PRINT("query oid: %08x"NEWLINE, Query->Oid);
#endif
            Status = NDIS_STATUS_NOT_SUPPORTED;
            break;
    }

    if (Status == NDIS_STATUS_SUCCESS)
    {
        if (pInfo == &qCounter)
        {
            if (Query->InformationBufferLength >= sizeof(ULONG64) ||
                Query->InformationBufferLength == 0)
            {
                ulInfoLen = sizeof(ULONG64);
            }
            else
            {
                ulInfoLen = sizeof(ULONG);
            }
            // We should always report that only 8 bytes are required to keep
            // ndistest happy
            Query->BytesNeeded = sizeof(ULONG64);
        }
        if (ulInfoLen <= Query->InformationBufferLength)
        {
            // Copy result into InformationBuffer
            if (ulInfoLen)
            {
                NdisMoveMemory(Query->InformationBuffer, pInfo, ulInfoLen);
                Query->BytesWritten = ulInfoLen;
            }
        }
        else
        {
            // too short
            Query->BytesNeeded = ulInfoLen;
            Status = NDIS_STATUS_BUFFER_TOO_SHORT;
        }
    }
    return Status;
}  // MiniportQueryInformation

/* -------------------------------------------------------------------------- */

/*
    CheckInputBufferSize

Description:
    This helper function checks the size of the buffer is big enough to be
    processed.

Parameters:
    UINT BufferSize
        The size of the buffer needed.

    struct _SET *Set
        The Ndis OID request set structure for storing the bytes needed.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS         Buffer is okay
    NDIS_STATUS_INVALID_LENGTH  Invalid input buffer data length
*/

static
NDIS_STATUS CheckInputBufferSize (
    UINT        BufferSize,
    struct _SET *Set )
{
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;

    if (Set->InformationBufferLength != BufferSize)
    {
        Set->BytesNeeded = BufferSize;
        Status = NDIS_STATUS_INVALID_LENGTH;
    }
    return Status;
}  // CheckInputBufferSize


/*
    MiniportSetInformation

Description:
    This function is used by NDIS to set adapter information by pre-defined
    NDIS_OIDs.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PNDIS_OID_REQUEST NdisSetRequest
        The NDIS_OID request to process.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS         Set Information was successful
    NDIS_STATUS_INVALID_LENGTH  Invalid input buffer data length
    NDIS_STATUS_INVALID_OID     Invalid OID
    NDIS_STATUS_NOT_SUPPORTED   Valid OID, operation not currently
                                supported
    NDIS_STATUS_INVALID_DATA    Invalid data in the input buffer
*/

NDIS_STATUS MiniportSetInformation (
    IN  NDIS_HANDLE       hAdapterContext,
    IN  PNDIS_OID_REQUEST NdisSetRequest )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;
    struct _SET *Set = &NdisSetRequest->DATA.SET_INFORMATION;
    PULONG pUlong = (PULONG) Set->InformationBuffer;

#ifdef DBG
    DbgPrintOidName("set", Set->Oid);
#endif

    Set->BytesNeeded = 0;
    Set->BytesRead = 0;
    switch (Set->Oid)
    {
        case OID_802_3_MULTICAST_LIST:
            //
            // Set the multicast address list on the NIC for packet reception.
            // The NIC driver can set a limit on the number of multicast
            // addresses bound protocol drivers can enable simultaneously.
            // NDIS returns NDIS_STATUS_MULTICAST_FULL if a protocol driver
            // exceeds this limit or if it specifies an invalid multicast
            // address.
            //
#ifdef DBG
            DBG_PRINT("mcast list len: %u\n", Set->InformationBufferLength);
#endif
            if ((Set->InformationBufferLength % MAC_ADDRESS_LENGTH) != 0)
            {
                Status = NDIS_STATUS_INVALID_LENGTH;
                break;
            }

            if ((Set->InformationBufferLength / MAC_ADDRESS_LENGTH) >
                    MAX_MULTICAST_LIST)
            {

#if 0
                Status = NDIS_STATUS_MULTICAST_FULL;
                Set->BytesNeeded = MAX_MULTICAST_LIST * MAC_ADDRESS_LENGTH;

#else
                /* Driver is being disabled.  Receive has been stopped. */
                if (NDIS_PAUSE_UNBIND_PROTOCOL == pAdapter->PauseReason)
                {
#ifdef DBG
                    DBG_PRINT("set multicast: driver being disabled\n");
#endif
                    break;
                }
                if (!AcquireAdapter(pAdapter, TRUE))
                {
                    Status = NDIS_STATUS_FAILURE;
                    break;
                }
                if (!HardwareSetMulticast(&pAdapter->Hardware, TRUE))
                {
                    Status = NDIS_STATUS_FAILURE;
                }
                ReleaseAdapter(pAdapter);
#endif
                break;
            }

            pAdapter->Hardware.m_bMulticastListSize = (UCHAR)
                (Set->InformationBufferLength / MAC_ADDRESS_LENGTH);
            NdisMoveMemory(pAdapter->Hardware.m_bMulticastList,
                Set->InformationBuffer, Set->InformationBufferLength);

            /* Driver is being disabled.  Receive has been stopped. */
            if (NDIS_PAUSE_UNBIND_PROTOCOL == pAdapter->PauseReason)
            {
#ifdef DBG
                DBG_PRINT("set multicast: driver being disabled\n");
#endif
                break;
            }
            if (!AcquireAdapter(pAdapter, TRUE))
            {
                Status = NDIS_STATUS_FAILURE;
                break;
            }
            if (!HardwareSetGroupAddress(&pAdapter->Hardware))
            {
                Status = NDIS_STATUS_FAILURE;
            }
            ReleaseAdapter(pAdapter);
            break;

        case OID_GEN_CURRENT_PACKET_FILTER:
        {
            //
            // Program the hardware to indicate the packets
            // of certain filter types.
            //
            UCHAR bPromiscuous;

            if (!NT_SUCCESS(Status = CheckInputBufferSize(sizeof(ULONG),
                    Set)))
            {
                break;
            }

#ifdef DBG
            DBG_PRINT("set filter:%08x"NEWLINE, pUlong[0]);
#endif

            if (pUlong[0] & pAdapter->ulPacketsNotAllowed)
            {
                Status = NDIS_STATUS_NOT_SUPPORTED;
                break;
            }

            pAdapter->ulPacketFilter = pUlong[0];

            /* Driver is being disabled.  Receive has been stopped. */
            if (NDIS_PAUSE_UNBIND_PROTOCOL == pAdapter->PauseReason)
            {
#ifdef DBG
                DBG_PRINT("set filter: driver being disabled\n");
#endif
                break;
            }

            if (!AcquireAdapter(pAdapter, TRUE))
            {
                Status = NDIS_STATUS_FAILURE;
                break;
            }
            if (pAdapter->ulPacketFilter)
            {
                // Now that the packet filter has been received, allow receive
                // packets to be indicated to the protocol stack.
                pAdapter->bPacketFilterSet = TRUE;

                /* Multicast receive is always enabled. */
                if ((pAdapter->ulPacketFilter &
                        NDIS_PACKET_TYPE_MULTICAST))
                {
                }
                if ((pAdapter->ulPacketFilter &
                        NDIS_PACKET_TYPE_ALL_MULTICAST))
                    bPromiscuous = TRUE;
                else
                    bPromiscuous = FALSE;
                if (!HardwareSetMulticast(&pAdapter->Hardware,
                        bPromiscuous))
                {
                    Status = NDIS_STATUS_FAILURE;
                }
                if ((pAdapter->ulPacketFilter &
                        (NDIS_PACKET_TYPE_PROMISCUOUS |
                        NDIS_PACKET_TYPE_ALL_LOCAL)))
                    bPromiscuous = TRUE;
                else
                    bPromiscuous = FALSE;
                if (!HardwareSetPromiscuous(&pAdapter->Hardware,
                        bPromiscuous))
                {
                    Status = NDIS_STATUS_FAILURE;
                }
            }
            else
            {
                pAdapter->bPacketFilterSet = FALSE;
                HardwareStopReceive(&pAdapter->Hardware);
            }
            ReleaseAdapter(pAdapter);
            break;
        }

        case OID_GEN_CURRENT_LOOKAHEAD:
            //
            // A protocol driver can set a suggested value for the number
            // of bytes to be used in its binding; however, the underlying
            // NIC driver is never required to limit its indications to
            // the value set.
            //
            if (!NT_SUCCESS(Status = CheckInputBufferSize(sizeof(ULONG),
                    Set)))
            {
                break;
            }

            if (pUlong[0] > NDIS_MAX_LOOKAHEAD)
            {
                Status = NDIS_STATUS_INVALID_DATA;
                break;
            }

#ifdef DBG
            DBG_PRINT("set max lookahead: %u, %u"NEWLINE,
                pAdapter->ulMaxLookAhead, pUlong[0]);
#endif
            if (pAdapter->ulMaxLookAhead < pUlong[0])
                pAdapter->ulMaxLookAhead = pUlong[0];
            break;

        case OID_GEN_MACHINE_NAME:
        {
#ifdef DBG
            DBG_PRINT("gen_machine_name"NEWLINE);
            if (Set->InformationBufferLength > 4)
            {
                PWCHAR pData = Set->InformationBuffer;
                ULONG i;

                for (i = 0; i < Set->InformationBufferLength / 2; i++)
                    DbgPrint("%c\n", pData[i]);
            }
#endif
            break;
        }

        case OID_802_3_CURRENT_ADDRESS:
            if (!NT_SUCCESS(Status = CheckInputBufferSize(MAC_ADDRESS_LENGTH,
                    Set)))
            {
                break;
            }

            NdisMoveMemory(pAdapter->Hardware.m_bPermanentAddress,
                Set->InformationBuffer, MAC_ADDRESS_LENGTH);
            break;

        case OID_PNP_SET_POWER:
            if (!NT_SUCCESS(Status = CheckInputBufferSize(
                    sizeof(NDIS_DEVICE_POWER_STATE), Set)))
            {
                break;
            }

            // D0 power state (full power) is the only state supported
            pAdapter->NextPowerState =
                *((PNDIS_DEVICE_POWER_STATE) Set->InformationBuffer);

#if DBG
DBG_PRINT("set power: %u, %u; %u, %u"NEWLINE, Set->InformationBufferLength,
    sizeof(NDIS_DEVICE_POWER_STATE), pAdapter->CurrentPowerState,
    pAdapter->NextPowerState);
#endif
            if (pAdapter->CurrentPowerState != pAdapter->NextPowerState)
            {
                switch (pAdapter->NextPowerState)
                {
                    case NdisDeviceStateD0:
                        DriverWake(pAdapter, pAdapter->CurrentPowerState);
                        break;
                    case NdisDeviceStateD1:
                    case NdisDeviceStateD2:
                    case NdisDeviceStateD3:
                        if (NdisDeviceStateD0 == pAdapter->CurrentPowerState)
                        {
                            DriverSleep(pAdapter);
                        }
                        break;
                }
                pAdapter->CurrentPowerState = pAdapter->NextPowerState;
            }
            break;

        case OID_PNP_ADD_WAKE_UP_PATTERN:

#if DBG_
DBG_PRINT("add wakeup: %u"NEWLINE, Set->InformationBufferLength);
#endif
            return AddWakeUpPattern(pAdapter,
                (PNDIS_PM_PACKET_PATTERN) Set->InformationBuffer,
                &Set->BytesRead);

        case OID_PNP_REMOVE_WAKE_UP_PATTERN:

#if DBG_
DBG_PRINT("remove wakeup: %u"NEWLINE, Set->InformationBufferLength);
#endif
            return RemoveWakeUpPattern(pAdapter,
                (PNDIS_PM_PACKET_PATTERN) Set->InformationBuffer,
                 &Set->BytesRead);

        case OID_PNP_ENABLE_WAKE_UP:
            if (!NT_SUCCESS(Status = CheckInputBufferSize(sizeof(ULONG),
                    Set)))
            {
                break;
            }

            pAdapter->ulWakeUpEnable = pUlong[0];

#if DBG
            if (pAdapter->ulWakeUpEnable & NDIS_PNP_WAKE_UP_MAGIC_PACKET)
                DBG_PRINT("NDIS_PNP_WAKE_UP_MAGIC_PNP"NEWLINE);
            if (pAdapter->ulWakeUpEnable & NDIS_PNP_WAKE_UP_PATTERN_MATCH)
                DBG_PRINT("NDIS_PNP_WAKE_UP_PATTERN_MATCH"NEWLINE);
            if (pAdapter->ulWakeUpEnable & NDIS_PNP_WAKE_UP_LINK_CHANGE)
                DBG_PRINT("NDIS_PNP_WAKE_UP_LINK_CHANGE"NEWLINE);
#endif
            if (pAdapter->ulWakeUpEnable & NDIS_PNP_WAKE_UP_MAGIC_PACKET)
                HardwareSetWolEnable(&pAdapter->Hardware, WOL_MAGIC_ENABLE,
                    TRUE);
            if (pAdapter->ulWakeUpEnable & NDIS_PNP_WAKE_UP_PATTERN_MATCH)
            {
                int nIndex;

                for (nIndex = 0; nIndex < MAX_WAKE_UP_PACKET; nIndex++)
                    if (pAdapter->WakeUpPacket[nIndex].
                            NdisPacketPattern.MaskSize)
                        HardwareSetWolEnable(&pAdapter->Hardware,
                            (USHORT)(1 << nIndex), TRUE);
            }
            break;

#if 0
        case OID_TCP_TASK_OFFLOAD:
            return SetTaskOffload(pAdapter,
                (PNDIS_TASK_OFFLOAD_HEADER) Set->InformationBuffer,
                &Set->BytesRead);
#endif

        case OID_GEN_NETWORK_LAYER_ADDRESSES:
        {
#if DBG
            PNETWORK_ADDRESS_LIST pAddrList =
                (PNETWORK_ADDRESS_LIST) Set->InformationBuffer;
            int i, j;

DBG_PRINT("network addr: %d"NEWLINE, pAddrList->AddressCount);
            for (i = 0; i < pAddrList->AddressCount; i++) {
DBG_PRINT("%d: ", pAddrList->Address[i].AddressType);
                for (j = 0; j < pAddrList->Address[i].AddressLength; j++) {
DBG_PRINT("%02X ", pAddrList->Address[i].Address[j]);
                }
DBG_PRINT(NEWLINE);
            }
#endif
            break;
        }

        case OID_GEN_TRANSPORT_HEADER_OFFSET:
        {
#if DBG
            PTRANSPORT_HEADER_OFFSET pOffset =
                (PTRANSPORT_HEADER_OFFSET) Set->InformationBuffer;
#endif

#if DBG
            if (NDIS_PROTOCOL_ID_TCP_IP == pOffset->ProtocolType)
            {
                DBG_PRINT("IP header offset: %d, %d"NEWLINE,
                    pOffset->ProtocolType, pOffset->HeaderOffset);
            }
#endif
            break;
        }

        case OID_DEVICE_MODE:
            Status = SetDeviceMode(pAdapter, Set->Oid,
                Set->InformationBuffer, Set->InformationBufferLength,
                &Set->BytesRead, &Set->BytesNeeded);
            break;

#if defined(NDIS620_MINIPORT)
        case OID_RECEIVE_FILTER_FREE_QUEUE:
            //
            // Free the requested receive queue.
            //
            break;

        case OID_RECEIVE_FILTER_CLEAR_FILTER:
            //
            // Remove the requested filter on the requested receive queue.
            //
            break;

        case OID_RECEIVE_FILTER_QUEUE_PARAMETERS:
            //
            // Update the queue information.
            //
            break;
#endif

        default:
#ifdef DBG
            DBG_PRINT("set oid: %08x"NEWLINE, Set->Oid);
#endif
            Status = NDIS_STATUS_INVALID_OID;
            break;
    }

    if (Status == NDIS_STATUS_SUCCESS)
    {
        if (!Set->BytesRead)
            Set->BytesRead = Set->InformationBufferLength;
    }

    return Status;
}


/*
    MiniportMethodRequest

Description:
    This function is used by NDIS to perform a WMI OID request.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PNDIS_OID_REQUEST NdisRequest
        The NDIS_OID request to process.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS         Request was successful
    NDIS_STATUS_NOT_SUPPORTED   Valid OID, operation not currently
                                supported
*/

NDIS_STATUS MiniportMethodRequest (
    IN  NDIS_HANDLE       hAdapterContext,
    IN  PNDIS_OID_REQUEST NdisRequest )
{
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;
    NDIS_OID Oid = NdisRequest->DATA.METHOD_INFORMATION.Oid;

    UNREFERENCED_PARAMETER(hAdapterContext);
    UNREFERENCED_PARAMETER(NdisRequest);

#ifdef DBG
    DbgPrintOidName("method", Oid);
#endif

    switch (Oid)
    {
#if defined(NDIS620_MINIPORT)
        case OID_RECEIVE_FILTER_ALLOCATE_QUEUE:
            //
            // Allocate the requested receive queue.
            //
            break;

        case OID_RECEIVE_FILTER_QUEUE_ALLOCATION_COMPLETE:
            //
            // Complete any remaining allocation for receive queues.
            //
            break;

        case OID_RECEIVE_FILTER_SET_FILTER:
            //
            // Add a filter to the requested queue.
            //
            break;
#endif

        default:
            Status = NDIS_STATUS_NOT_SUPPORTED;
            break;
    }

    return Status;
}


/*
    MiniportCancelOidRequest

Description:
    This procedure is used by NDIS to abort an asynchronous OID request.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PVOID RequestId
        An identifier that corresponds to the RequestId field of the
        NDIS_OID_REQUEST.

Return (None):
*/

VOID MiniportCancelOidRequest (
    IN  NDIS_HANDLE hAdapterContext,
    IN  PVOID       RequestId )
{
    UNREFERENCED_PARAMETER(hAdapterContext);
    UNREFERENCED_PARAMETER(RequestId);

    //
    // This miniport sample does not pend any OID requests, so we don't have
    // to worry about cancelling them.
    //
}  // MiniportCancelOidRequest


/*
    MiniportOidRequest

Description:
    This function is used by NDIS to get or set the value of a specified OID.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PNDIS_OID_REQUEST NdisRequest
        The NDIS_OID request to process.

Return (NDIS_STATUS):
    Return code from one of the NdisRequest subfunctions.
*/

NDIS_STATUS MiniportOidRequest (
    IN  NDIS_HANDLE       hAdapterContext,
    IN  PNDIS_OID_REQUEST NdisRequest )
{
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;

    switch (NdisRequest->RequestType)
    {
        case NdisRequestMethod:
            Status = MiniportMethodRequest(hAdapterContext, NdisRequest);
            break;

        case NdisRequestSetInformation:
            Status = MiniportSetInformation(hAdapterContext, NdisRequest);
            break;

        case NdisRequestQueryInformation:
        case NdisRequestQueryStatistics:
            Status = MiniportQueryInformation(hAdapterContext, NdisRequest);
            break;

        default:
            //
            // The entry point may by used by other requests
            //
            Status = NDIS_STATUS_NOT_SUPPORTED;
            break;
    }
    return Status;
}  // MiniportOidRequest
